"""
Text Module
===========

This file defines general incremental units and incremental modules that deal with text.
This may be a transcription that is generated from an ASR module, a text IU containing
words to be synthesized by a TTS module or other general purpose text.
"""

import retico_core.core.abstract as retico_core


def get_text_increment(module, new_text):
    """Compares the full text given by the asr with the IUs that are already
    produced (current_output) and returns only the increment from the last update. It
    revokes all previously produced IUs that do not match.

    For example, if the ``current_output`` of the module consists of the follwing IUs:
        [The] [quick] [bright]
    and the ``new_text`` is:
        "The quick brown fox"
    The function would return a new UpdateMessage
        UpdateMessage([bright], REVOKE)
    and the follwing new tokens:
        ["brown", "fox"]

    This function does not create UPDATE messages (e.g., updating [bright] to [brown] in
    the example above), but rather REVOKEs all unfitting IUs.

    Args:
        module (IncrementalModule): The incremental module that produces ``TextIU`` s
        new_text (str): The full new text that should be incrementalized

    Returns:
        UpdateMessage, list: Returns an update message containing the IUs that were
            revoked (with the REVOKE update type) and a list of new tokens (strings) for
            which new IUs should be created.
    """
    um = retico_core.UpdateMessage()
    tokens = new_text.strip().split(" ")
    if tokens == [""]:
        return um, []

    new_tokens = []
    iu_idx = 0
    token_idx = 0
    while token_idx < len(tokens):
        if iu_idx >= len(module.current_output):
            new_tokens.append(tokens[token_idx])
            token_idx += 1
        else:
            current_iu = module.current_output[iu_idx]
            iu_idx += 1
            if tokens[token_idx] == current_iu.text:
                token_idx += 1
            else:
                current_iu.revoked = True
                um.add_iu(current_iu, retico_core.UpdateType.REVOKE)
    module.current_output = [iu for iu in module.current_output if not iu.revoked]

    return um, new_tokens


class TextIU(retico_core.IncrementalUnit):
    """An IU that contains text."""

    @staticmethod
    def type():
        return "Text IU"

    def get_text(self):
        """Return the text contained in the IU.

        Returns:
            str: The text contained in the IU.
        """
        return self.payload

    def set_text(self, text):
        """Sets the text contained in the IU.

        Args:
            text (str): The new text of the IU
        """
        self.payload = text

    text = property(get_text, set_text)


class GeneratedTextIU(TextIU):
    """An IU that contains generated text.

    This includes information about whether the text should be dispatched once
    it has been transformed into speech."""

    @staticmethod
    def type():
        return "Generated Text IU"

    def __init__(self, dispatch=False, **kwargs):
        super().__init__(**kwargs)
        self.dispatch = dispatch


class SpeechRecognitionIU(TextIU):
    """An IU that contains information about recognized speech."""

    @staticmethod
    def type():
        return "Speech Recgonition IU"

    def __init__(
        self, creator, iuid=0, previous_iu=None, grounded_in=None, payload=None
    ):
        super().__init__(
            creator,
            iuid=iuid,
            previous_iu=previous_iu,
            grounded_in=grounded_in,
            payload=payload,
        )
        self.predictions = None
        self.stability = None
        self.confidence = None
        self.payload = payload
        self.text = None
        self.final = False

    def set_asr_results(self, predictions, text, stability, confidence, final):
        """Set the asr results for the SpeechRecognitionIU.

        Args:
            predictions (list): A list of predictions. This will also set the
                payload. The last prediction in this list should be the latest
                and best prediction.
            text (str): The text of the latest prediction
            stability (float): The stability of the latest prediction
            confidence (float): The confidence in the latest prediction
            final (boolean): Whether the prediction is final
        """
        self.predictions = predictions
        self.payload = predictions
        self.text = text
        self.stability = stability
        self.confidence = confidence
        self.final = final

    def get_text(self):
        return self.text


class TextRecorderModule(retico_core.AbstractConsumingModule):
    """A module that writes the received text into a file."""

    @staticmethod
    def name():
        return "Text Recorder Module"

    @staticmethod
    def description():
        return "A module that writes received TextIUs to file"

    @staticmethod
    def input_ius():
        return [TextIU, GeneratedTextIU, SpeechRecognitionIU]

    def __init__(self, filename, separator="\t", **kwargs):
        super().__init__(**kwargs)
        self.filename = filename
        self.separator = separator
        self.txt_file = None

    def setup(self):
        self.txt_file = open(self.filename, "w")

    def shutdown(self):
        if self.txt_file:
            self.txt_file.close()
            self.txt_file = None

    def process_update(self, update_message):
        for iu, ut in update_message:
            if ut != retico_core.UpdateType.ADD:
                continue
            if self.txt_file:
                self.txt_file.write(str(iu.grounded_in.creator))
                self.txt_file.write(self.separator)
                self.txt_file.write(str(iu.created_at))
                self.txt_file.write(self.separator)
                self.txt_file.write(iu.get_text())
                if isinstance(iu, GeneratedTextIU):
                    self.txt_file.write(self.separator)
                    self.txt_file.write(str(iu.dispatch))
                if isinstance(iu, SpeechRecognitionIU):
                    self.txt_file.write(self.separator)
                    self.txt_file.write(str(iu.predictions))
                    self.txt_file.write(self.separator)
                    self.txt_file.write(str(iu.stability))
                    self.txt_file.write(self.separator)
                    self.txt_file.write(str(iu.confidence))
                    self.txt_file.write(self.separator)
                    self.txt_file.write(str(iu.final))
                self.txt_file.write("\n")


class TextTriggerModule(retico_core.AbstractTriggerModule):
    """A trigger module that creates a TextIU once its trigger function is called."""

    @staticmethod
    def name():
        return "Text Trigger Module"

    @staticmethod
    def description():
        return "A trigger module that creates a TextIU once its triggered"

    @staticmethod
    def output_iu():
        return GeneratedTextIU

    def __init__(self, dispatch=True, **kwargs):
        super().__init__(**kwargs)
        self.dispatch = dispatch

    def trigger(self, data={}, update_type=retico_core.UpdateType.ADD):
        text = data.get("text", "This is a trigger test")
        output_iu = self.create_iu()
        output_iu.payload = text
        output_iu.dispatch = self.dispatch
        self.append(retico_core.UpdateMessage.from_iu(output_iu, update_type))


class TextDispatcherModule(retico_core.AbstractModule):
    """
    A Moduel that turns SpeechRecognitionIUs or TextIUs into GeneratedTextIUs
    that have the dispatch-flag set.
    """

    @staticmethod
    def name():
        return "ASR to TTS Module"

    @staticmethod
    def description():
        return "A module that uses SpeechRecognition IUs and outputs dispatchable IUs"

    @staticmethod
    def input_ius():
        return [SpeechRecognitionIU, TextIU]

    @staticmethod
    def output_iu():
        return GeneratedTextIU

    def __init__(self, dispatch_final=True, **kwargs):
        super().__init__(**kwargs)
        self.dispatch_final = dispatch_final

    def process_update(self, update_message):
        um = retico_core.UpdateMessage()
        for iu, ut in update_message:
            output_iu = self.create_iu(iu)
            output_iu.payload = iu.get_text()
            output_iu.dispatch = True
            if isinstance(iu, SpeechRecognitionIU) and self.dispatch_final:
                output_iu.dispatch = iu.final
            output_iu.committed = iu.committed
            um.add_iu(output_iu, ut)
        return um


class IncrementalizeASRModule(retico_core.AbstractModule):
    """A module that takes the output of a non-incremental ASR module, where each IU
    contains the full text of the speech recognition and produces increments based on
    the difference to the last output.
    """

    @staticmethod
    def name():
        return "Incrementalize ASR Module"

    @staticmethod
    def description():
        return (
            "A module that takes SpeechRecognitionIUs and emits only the "
            + "increments from the previous iu"
        )

    @staticmethod
    def input_ius():
        return [SpeechRecognitionIU]

    @staticmethod
    def output_iu():
        return SpeechRecognitionIU

    def __init__(self, threshold=0.8, **kwargs):
        super().__init__(**kwargs)
        self.threshold = threshold

    def process_update(self, update_message):
        um = retico_core.UpdateMessage()
        for iu, ut in update_message:
            if ut != retico_core.UpdateType.ADD:
                continue
            if iu.stability < self.threshold and iu.confidence == 0.0:
                continue
            current_text = iu.get_text()
            if self.current_input:
                um, current_text = retico_core.text.get_text_increment(
                    self, current_text
                )
            if len(current_text) == 0:
                continue

            output_iu = self.create_iu(iu)

            # Just copy the input IU
            output_iu.set_asr_results(
                iu.predictions,
                current_text,
                iu.stability,
                iu.confidence,
                iu.final,
            )
            self.current_input.append(output_iu)

            if output_iu.final:
                self.current_input = []
                output_iu.committed = True

            um.add_iu(output_iu, retico_core.UpdateType.ADD)

        return um


class EndOfUtteranceModule(retico_core.AbstractModule):
    """A module that looks for the "final" flag of a SpeechRecognitionIU and forwards
    an EndOfTurnIU when the SpeechRecognition detected that the utterance is finished.
    """

    @staticmethod
    def name():
        return "End of Utterance Module"

    @staticmethod
    def description():
        return "A module that forwards the end of utterance from the ASR output"

    @staticmethod
    def input_ius():
        return [SpeechRecognitionIU]

    @staticmethod
    def output_iu():
        return retico_core.dialogue.EndOfTurnIU

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def process_update(self, update_message):
        um = retico_core.UpdateMessage()
        for iu, ut in update_message:
            if ut == retico_core.UpdateType.ADD:
                if iu.final:
                    outiu = self.create_iu(iu)
                    outiu.set_eot(1.0, False)
                    um.add_iu(outiu, retico_core.UpdateType.ADD)
        return um



"""
Audio Module
============

This module defines basic incremental units and incremental modules to handle
audio input (via a standard microphone) and output.
"""

import threading
import queue
import time
import wave
import platform

import pyaudio

import retico_core.core.abstract as retico


CHANNELS = 1
"""Number of channels. For now, this is hard coded MONO. If there is interest to do
stereo or audio with even more channels, it has to be integrated into the modules."""

TIMEOUT = 0.01
"""Timeout in seconds used for the StreamingSpeakerModule."""


def show_audio_devices():
    """Shows all availbale audio input and output devices using pyAudio."""
    p = pyaudio.PyAudio()

    info = p.get_host_api_info_by_index(0)

    print("Output Devices:")
    for i in range(info["deviceCount"]):
        device = p.get_device_info_by_host_api_device_index(0, i)
        if device["maxOutputChannels"] > 0:
            print("  %s (%d)" % (device["name"], device["index"]))

    print("\nInput Devices:")
    for i in range(info["deviceCount"]):
        device = p.get_device_info_by_host_api_device_index(0, i)
        if device["maxInputChannels"] > 0:
            print("  %s (%d)" % (device["name"], device["index"]))


class AudioIU(retico.IncrementalUnit):
    """An audio incremental unit that receives raw audio data from a source.

    The audio contained should be monaural.

    Attributes:
        creator (AbstractModule): The module that created this IU
        previous_iu (IncrementalUnit): A link to the IU created before the
            current one.
        grounded_in (IncrementalUnit): A link to the IU this IU is based on.
        created_at (float): The UNIX timestamp of the moment the IU is created.
        raw_audio (bytes[]): The raw audio of this IU
        rate (int): The frame rate of this IU
        nframes (int): The number of frames of this IU
        sample_width (int): The bytes per sample of this IU
    """

    @staticmethod
    def type():
        return "Audio IU"

    def __init__(
        self,
        creator=None,
        iuid=0,
        previous_iu=None,
        grounded_in=None,
        rate=None,
        nframes=None,
        sample_width=None,
        raw_audio=None,
        **kwargs
    ):
        super().__init__(
            creator=creator,
            iuid=iuid,
            previous_iu=previous_iu,
            grounded_in=grounded_in,
            payload=raw_audio,
        )
        self.raw_audio = raw_audio
        self.rate = rate
        self.nframes = nframes
        self.sample_width = sample_width

    def set_audio(self, raw_audio, nframes, rate, sample_width):
        """Sets the audio content of the IU."""
        self.raw_audio = raw_audio
        self.payload = raw_audio
        self.nframes = int(nframes)
        self.rate = int(rate)
        self.sample_width = int(sample_width)

    def audio_length(self):
        """Return the length of the audio IU in seconds.

        Returns:
            float: Length of the audio in this IU in seconds.
        """
        return float(self.nframes) / float(self.rate)


class SpeechIU(AudioIU):
    """A type of audio incremental unit that contains a larger amount of audio
    information and the information if the audio should be dispatched or not.

    This IU can be processed by an AudioDispatcherModule which converts this
    type of IU to AudioIU.
    """

    @staticmethod
    def type():
        return "Speech IU"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.dispatch = False


class DispatchedAudioIU(AudioIU):
    """A type of audio incremental unit that is dispatched by an
    AudioDispatcherModule. It has the information of the percentual completion
    of the dispatched audio. This may be useful for a dialog manager that
    wants to track the status of the current dispatched audio.
    """

    @staticmethod
    def type():
        return "Dispatched Audio IU"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.completion = 0.0
        self.is_dispatching = False

    def set_dispatching(self, completion, is_dispatching):
        """Set the completion percentage and the is_dispatching flag.

        Args:
            completion (float): The degree of completion of the current
                utterance.
            is_dispatching (bool): Whether or not the dispatcher is currently
                dispatching
        """
        self.completion = completion
        self.is_dispatching = is_dispatching


class MicrophoneModule(retico.AbstractProducingModule):
    """A module that produces IUs containing audio signals that are captures by
    a microphone."""

    @staticmethod
    def name():
        return "Microphone Module"

    @staticmethod
    def description():
        return "A prodicing module that records audio from microphone."

    @staticmethod
    def output_iu():
        return AudioIU

    def callback(self, in_data, frame_count, time_info, status):
        """The callback function that gets called by pyaudio.

        Args:
            in_data (bytes[]): The raw audio that is coming in from the
                microphone
            frame_count (int): The number of frames that are stored in in_data
        """
        self.audio_buffer.put(in_data)
        return (in_data, pyaudio.paContinue)

    def __init__(self, frame_length=0.02, rate=44100, sample_width=2, **kwargs):
        """
        Initialize the Microphone Module.

        Args:
            frame_length (float): The length of one frame (i.e., IU) in seconds
            rate (int): The frame rate of the recording
            sample_width (int): The width of a single sample of audio in bytes.
        """
        super().__init__(**kwargs)
        self.frame_length = frame_length
        self.chunk_size = round(rate * frame_length)
        self.rate = rate
        self.sample_width = sample_width

        self._p = pyaudio.PyAudio()

        self.audio_buffer = queue.Queue()
        self.stream = None

    def process_update(self, _):
        if not self.audio_buffer:
            return None
        try:
            sample = self.audio_buffer.get()
        except queue.Empty:
            return None
        output_iu = self.create_iu()
        output_iu.set_audio(sample, self.chunk_size, self.rate, self.sample_width)
        return retico.UpdateMessage.from_iu(output_iu, retico.UpdateType.ADD)

    def setup(self):
        """Set up the microphone for recording."""
        p = self._p
        self.stream = p.open(
            format=p.get_format_from_width(self.sample_width),
            channels=CHANNELS,
            rate=self.rate,
            input=True,
            output=False,
            stream_callback=self.callback,
            frames_per_buffer=self.chunk_size,
            start=False,
        )

    def prepare_run(self):
        if self.stream:
            self.stream.start_stream()

    def shutdown(self):
        """Close the audio stream."""
        self.stream.stop_stream()
        self.stream.close()
        self.stream = None
        self.audio_buffer = queue.Queue()


class SpeakerModule(retico.AbstractConsumingModule):
    """A module that consumes AudioIUs of arbitrary size and outputs them to the
    speakers of the machine. When a new IU is incoming, the module blocks as
    long as the current IU is being played."""

    @staticmethod
    def name():
        return "Speaker Module"

    @staticmethod
    def description():
        return "A consuming module that plays audio from speakers."

    @staticmethod
    def input_ius():
        return [AudioIU]

    @staticmethod
    def output_iu():
        return None

    def __init__(
        self,
        rate=44100,
        sample_width=2,
        use_speaker="both",
        device_index=None,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.rate = rate
        self.sample_width = sample_width
        self.use_speaker = use_speaker

        self._p = pyaudio.PyAudio()

        if device_index is None:
            device_index = self._p.get_default_output_device_info()["index"]
        self.device_index = device_index

        self.stream = None
        self.time = None

    def process_update(self, update_message):
        for iu, ut in update_message:
            if ut == retico.UpdateType.ADD:
                self.stream.write(bytes(iu.raw_audio))
        return None

    def setup(self):
        """Set up the speaker for outputting audio"""
        p = self._p

        if platform.system() == "Darwin":
            if self.use_speaker == "left":
                stream_info = pyaudio.PaMacCoreStreamInfo(channel_map=(0, -1))
            elif self.use_speaker == "right":
                stream_info = pyaudio.PaMacCoreStreamInfo(channel_map=(-1, 0))
            else:
                stream_info = pyaudio.PaMacCoreStreamInfo(channel_map=(0, 0))
        else:
            stream_info = None

        self.stream = p.open(
            format=p.get_format_from_width(self.sample_width),
            channels=CHANNELS,
            rate=self.rate,
            input=False,
            output_host_api_specific_stream_info=stream_info,
            output=True,
            output_device_index=self.device_index,
        )

    def shutdown(self):
        """Close the audio stream."""
        self.stream.stop_stream()
        self.stream.close()
        self.stream = None


class StreamingSpeakerModule(retico.AbstractConsumingModule):
    """A module that consumes Audio IUs and outputs them to the speaker of the
    machine. The audio output is streamed and thus the Audio IUs have to have
    exactly [chunk_size] samples."""

    @staticmethod
    def name():
        return "Streaming Speaker Module"

    @staticmethod
    def description():
        return "A consuming module that plays audio from speakers."

    @staticmethod
    def input_ius():
        return [AudioIU]

    @staticmethod
    def output_iu():
        return None

    def callback(self, in_data, frame_count, time_info, status):
        """The callback function that gets called by pyaudio."""
        if self.audio_buffer:
            try:
                audio_paket = self.audio_buffer.get(timeout=TIMEOUT)
                return (audio_paket, pyaudio.paContinue)
            except queue.Empty:
                pass
        return (b"\0" * frame_count * self.sample_width, pyaudio.paContinue)

    def __init__(self, frame_length=0.02, rate=44100, sample_width=2, **kwargs):
        """Initialize the streaming speaker module.

        Args:
            frame_length (float): The length of one frame (i.e., IU) in seconds.
            rate (int): The frame rate of the audio. Defaults to 44100.
            sample_width (int): The sample width of the audio. Defaults to 2.
        """
        super().__init__(**kwargs)
        self.frame_length = frame_length
        self.chunk_size = round(rate * frame_length)
        self.rate = rate
        self.sample_width = sample_width

        self._p = pyaudio.PyAudio()

        self.audio_buffer = queue.Queue()
        self.stream = None

    def process_update(self, update_message):
        for iu, ut in update_message:
            if ut == retico.UpdateType.ADD:
                self.audio_buffer.put(iu.raw_audio)
        return None

    def setup(self):
        """Set up the speaker for speaking...?"""
        p = self._p
        self.stream = p.open(
            format=p.get_format_from_width(self.sample_width),
            channels=CHANNELS,
            rate=self.rate,
            input=False,
            output=True,
            stream_callback=self.callback,
            frames_per_buffer=self.chunk_size,
        )

    def prepare_run(self):
        self.stream.start_stream()

    def shutdown(self):
        """Close the audio stream."""
        self.stream.stop_stream()
        self.stream.close()
        self.stream = None
        self.audio_buffer = queue.Queue()


class AudioDispatcherModule(retico.AbstractModule):
    """An Audio module that takes a raw audio stream of arbitrary size and
    outputs AudioIUs with a specific chunk size at the rate it would be produced
    if the audio was being played.

    This could be espacially useful when an agents' TTS module produces an
    utterance, but this utterance should not be transmitted as a whole but in
    an incremental way.

    Attributes:
        target_frame_length (float): The size of each output IU in seconds.
        target_chunk_size (int): The size of each output IU in samples.
        silence (bytes): A bytes array containing [target_chunk_size] samples
            of silence that is dispatched when [continuous] is True and no input
            IU is dispatched.
        continuous (bool): Whether or not the dispatching should be continuous.
            If True, AudioIUs with "silence" will be disptached if no input IUs
            are being dispatched. If False, no IUs will be produced during
            silence.
        rate (int): The sample rate of the outout and the input IU.
        sample_width (int): The sample with of the output and input IU.
        speed (float): The speed of the dispatching. 1.0 means realtime.
        dispatching_mutex (threading.Lock): The mutex if an input IU is
            currently being dispatched.
        audio_buffer (list): The current audio buffer containing the output IUs
            that are currently dispatched.
        run_loop (bool): Whether or not the dispatching loop is running.
        interrupt (bool): Whether or not incoming IUs interrupt the old
            dispatching
    """

    @staticmethod
    def name():
        return "Audio Dispatching Module"

    @staticmethod
    def description():
        return (
            "A module that transmits audio by splitting it up into" "streamable pakets."
        )

    @staticmethod
    def input_ius():
        return [SpeechIU]

    @staticmethod
    def output_iu():
        return DispatchedAudioIU

    def __init__(
        self,
        target_frame_length=0.02,
        rate=44100,
        sample_width=2,
        speed=1.0,
        continuous=True,
        silence=None,
        interrupt=True,
        **kwargs
    ):
        """Initialize the AudioDispatcherModule with the given arguments.

        Args:
            target_frame_length (float): The length of each output IU in seconds.
            rate (int): The sample rate of the outout and the input IU.
            sample_width (int): The sample with of the output and input IU.
            speed (float): The speed of the dispatching. 1.0 means realtime.
            continuous (bool): Whether or not the dispatching should be
                continuous. If True, AudioIUs with "silence" will be dispatched
                if no input IUs are being dispatched. If False, no IUs will be
                produced during silence.
            silence (bytes): A bytes array containing target_frame_length seconds of
                of silence. If this argument is set to None, a default silence
                of all zeros will be set.
            interrupt (boolean): If this flag is set, a new input IU with audio
                to dispatch will stop the current dispatching process. If set to
                False, the "old" dispatching will be finished before the new one
                is started. If the new input IU has the dispatching flag set to
                False, dispatching will always be stopped.
        """
        super().__init__(**kwargs)
        self.target_frame_length = target_frame_length
        self.target_chunk_size = round(target_frame_length * rate)
        if not silence:
            self.silence = b"\0" * self.target_chunk_size * sample_width
        else:
            self.silence = silence
        self.continuous = continuous
        self.rate = rate
        self.sample_width = sample_width
        self._is_dispatching = False
        self.dispatching_mutex = threading.Lock()
        self.audio_buffer = []
        self.run_loop = False
        self.speed = speed
        self.interrupt = interrupt

    def is_dispatching(self):
        """Return whether or not the audio dispatcher is dispatching a Speech
        IU.

        Returns:
            bool: Whether or not speech is currently dispatched
        """
        with self.dispatching_mutex:
            return self._is_dispatching

    def set_dispatching(self, value):
        """Set the dispatching value of this module in a thread safe way.

        Args:
            value (bool): The new value of the dispatching flag.
        """
        with self.dispatching_mutex:
            self._is_dispatching = value

    def process_update(self, update_message):
        cur_width = self.target_chunk_size * self.sample_width
        # If the AudioDispatcherModule is set to intterupt mode or if the
        # incoming IU is set to not dispatch, we stop dispatching and clean the
        # buffer
        for iu, ut in update_message:
            if ut != retico.UpdateType.ADD:
                continue
            if self.interrupt or not iu.dispatch:
                self.set_dispatching(False)
                self.audio_buffer = []
            if iu.dispatch:
                # Loop over all frames (frame-sized chunks of data) in the input IU
                # and add them to the buffer to be dispatched by the
                # _dispatch_audio_loop
                for i in range(0, iu.nframes, self.target_chunk_size):
                    cur_pos = i * self.sample_width
                    data = iu.raw_audio[cur_pos : cur_pos + cur_width]
                    distance = cur_width - len(data)
                    data += b"\0" * distance

                    completion = float((i + self.target_chunk_size) / iu.nframes)
                    if completion > 1:
                        completion = 1

                    current_iu = self.create_iu(iu)
                    current_iu.set_dispatching(completion, True)
                    current_iu.set_audio(
                        data, self.target_chunk_size, self.rate, self.sample_width
                    )
                    self.audio_buffer.append(current_iu)
                self.set_dispatching(True)
        return None

    def _dispatch_audio_loop(self):
        """A method run in a thread that adds IU to the output queue."""
        while self.run_loop:
            with self.dispatching_mutex:
                if self._is_dispatching:
                    if self.audio_buffer:
                        self.append(
                            retico.UpdateMessage.from_iu(
                                self.audio_buffer.pop(0), retico.UpdateType.ADD
                            )
                        )
                    else:
                        self._is_dispatching = False
                if not self._is_dispatching:  # no else here! bc line above
                    if self.continuous:
                        current_iu = self.create_iu(None)
                        current_iu.set_audio(
                            self.silence,
                            self.target_chunk_size,
                            self.rate,
                            self.sample_width,
                        )
                        current_iu.set_dispatching(0.0, False)
                        self.append(
                            retico.UpdateMessage.from_iu(
                                current_iu, retico.UpdateType.ADD
                            )
                        )
            time.sleep((self.target_chunk_size / self.rate) / self.speed)

    def prepare_run(self):
        self.run_loop = True
        t = threading.Thread(target=self._dispatch_audio_loop)
        t.start()

    def shutdown(self):
        self.run_loop = False
        self.audio_buffer = []


class AudioRecorderModule(retico.AbstractConsumingModule):
    """A Module that consumes AudioIUs and saves them as a PCM wave file to
    disk."""

    @staticmethod
    def name():
        return "Audio Recorder Module"

    @staticmethod
    def description():
        return "A Module that saves incoming audio to disk."

    @staticmethod
    def input_ius():
        return [AudioIU]

    def __init__(self, filename, rate=44100, sample_width=2, **kwargs):
        """Initialize the audio recorder module.

        Args:
            filename (string): The file name where the audio should be recorded
                to. The path to the file has to be created beforehand.
            rate (int): The sample rate of the input and thus of the wave file.
                Defaults to 44100.
            sample_width (int): The width of one sample. Defaults to 2.
        """
        super().__init__(**kwargs)
        self.filename = filename
        self.wavfile = None
        self.rate = rate
        self.sample_width = sample_width

    def process_update(self, update_message):
        for iu, ut in update_message:
            if ut == retico.UpdateType.ADD:
                self.wavfile.writeframes(iu.raw_audio)

    def setup(self):
        self.wavfile = wave.open(self.filename, "wb")
        self.wavfile.setframerate(self.rate)
        self.wavfile.setnchannels(CHANNELS)
        self.wavfile.setsampwidth(self.sample_width)

    def shutdown(self):
        self.wavfile.close()



"""
Abstract Module
===============

This module defines the abstract classes used by the incremental modules.
The AbstractModules defines some methods that handle the general tasks of a
module (like the handling of Queues).
The IncrementalQueue defines the basic functionality of an incremental queue
like appending an IU to the queue and letting modules subscribe to the Queue.
The Incremental Unit provides the basic data structure to exchange information
between modules.
"""

import queue
import threading
import time
import enum
import copy


class UpdateType(enum.Enum):
    """The update type enum that defines all the types with which the incremental units
    can be transmitted. Per default, the UpdateMessge class checks that the update type
    is one of the types listed in this enum. However, the strict type checking can be
    disabled and any update type may be used."""

    ADD = "add"
    UPDATE = "update"
    REVOKE = "revoke"
    COMMIT = "commit"


class IncrementalQueue(queue.Queue):
    """An abstract incremental queue.

    A module may subscribe to a queue of another module. Every time a new
    incremental unit (IU) is produced, the IU is put into a special queue for
    every subscriber to the incremental queue. Every unit gets its own queue and
    may process the items at different speeds.

    Attributes:
        provider (AbstractModule): The module that provides IUs for this queue.
        consumer (AbstractModule): The module that consumes IUs for this queue.
        maxsize (int): The maximum size of the queue, where 0 does not restrict
            the size.
    """

    def __init__(self, provider, consumer, maxsize=0):
        super().__init__(maxsize=maxsize)
        self.provider = provider
        self.consumer = consumer

    def remove(self):
        """Removes the queue from the consumer and the producer."""
        self.provider.remove_right_buffer(self)
        self.consumer.remove_left_buffer(self)


class IncrementalUnit:
    """An abstract incremental unit.

    The IU may be used for ASR, NLU, DM, TT, TTS, ... It can be redefined to fit
    the needs of the different module (and module-types) but should always
    provide these functionalities.

    The meta_data may be used when an incremental module is having additional
    information because it is working in a simulated environemnt. This data can
    be used by later modules to keep the simulation going.

    Attributes:
        creator (AbstractModule): The module that created this IU
        previous_iu (IncrementalUnit): A link to the IU created before the
            current one.
        grounded_in (IncrementalUnit): A link to the IU this IU is based on.
        created_at (float): The UNIX timestamp of the moment the IU is created.
        meta_data (dict): Meta data that offers optional meta information. This
            field can be used to add information that is not available for all
            uses of the specific incremental unit.
    """

    MAX_DEPTH = 50
    """Maximum depth of the previous_iu and grounded_in connections."""

    def __init__(
        self,
        creator=None,
        iuid=None,
        previous_iu=None,
        grounded_in=None,
        payload=None,
        **kwargs,
    ):
        """Initialize an abstract IU. Takes the module that created the IU as an
        argument.

        Args:
            creator (AbstractModule): The module that created this incremental
                unit.
            iuid (int): The id of the IU. This should be a unique ID given by the module
                that produces the incremental unit and is used to identify the IU later
                on - for example when revoking an IU.
            previous_iu (IncrementalUnit): A link to the incremental unit
                created before the current one by the same module.
            grounded_in (IncrementalUnit): A link to the incremental unit that
                this one is based on.
            payload: A generic payload that can be set.
        """
        self.creator = creator
        self.iuid = iuid
        if self.iuid is None:
            self.iuid = hash(self)
        self.previous_iu = previous_iu
        self.grounded_in = grounded_in
        self._processed_list = []
        self.payload = payload
        self.mutex = threading.Lock()

        self.committed = False
        self.revoked = False

        self.meta_data = {}
        if grounded_in:
            self.meta_data = {**grounded_in.meta_data}

        self.created_at = time.time()
        self._remove_old_links()

    def _remove_old_links(self):
        current_depth = 0
        previous_iu = self.previous_iu
        while previous_iu:
            if current_depth == self.MAX_DEPTH:
                previous_iu.previous_iu = None
            previous_iu = previous_iu.previous_iu
            current_depth += 1
        current_depth = 0
        grounded_in = self.grounded_in
        while grounded_in:
            if current_depth == self.MAX_DEPTH:
                grounded_in.grounded_in = None
            grounded_in = grounded_in.grounded_in
            current_depth += 1

    def age(self):
        """Returns the age of the IU in seconds.

        Returns:
            float: The age of the IU in seconds
        """
        return time.time() - self.created_at

    def older_than(self, s):
        """Return whether the IU is older than s seconds.

        Args:
            s (float): The time in seconds to check against.

        Returns:
            bool: Whether or not the age of the IU exceeds s seconds.
        """
        return self.age() > s

    def processed_list(self):
        """Return a list of all modules that have already processed this IU.

        The returned list is a copy of the list held by the IU.

        Returns:
            list: A list of all modules that have alread processed this IU.
        """
        with self.mutex:
            return list(self._processed_list)

    def set_processed(self, module):
        """Add the module to the list of modules that have already processed
        this IU.

        Args:
            module (AbstractModule): The module that has processed this IU.
        """
        if not isinstance(module, AbstractModule):
            raise TypeError("Given object is not a module!")
        with self.mutex:
            self._processed_list.append(module)

    def is_processed_by(self, module):
        """Return True if the IU is processed by the given module.

        If the given object is a module that has not processed this IU or is not
        a module it returns False.

        Args:
            module (AbstractModule): The module to test whether or not it has
                processed the IU

        Returns:
            bool: Whether or not the module has processed the IU.
        """
        with self.mutex:
            return module in self._processed_list

    def __repr__(self):
        return "%s - (%s): %s" % (
            self.type(),
            self.creator.name(),
            str(self.payload)[0:10],
        )

    def __eq__(self, other):
        if not isinstance(other, IncrementalUnit):
            return False
        return self.iuid == other.iuid

    @staticmethod
    def type():
        """Return the type of the IU in a human-readable format.

        Returns:
            str: The type of the IU in a human-readable format.
        """
        raise NotImplementedError()


class UpdateMessage:
    """A class that encapsulates multiple incremental units and their update type. The
    update types can be any of the ones defined in the enum UpdateType"""

    def __init__(self):
        """Initializes the update message with no IU added.

        To initialize with a single IU use the classmethod "from_iu" or for a list of
        IUs use the classmethod "from_ius".
        """
        self._msgs = []  # First element of tuple is IU, second is UpdateType
        self._counter = -1

    def __len__(self):
        return len(self._msgs)

    @classmethod
    def from_iu(cls, iu, update_type):
        """Initializes the update message with an initial pair of incremental unit and
        update type.

        Args:
            iu (IncrementalUnit): The first incremental unit of the update message
            update_type (UpdateType): The update type of the incremental unit.
        """
        um = UpdateMessage()
        um.add_iu(iu, update_type)
        return um

    @classmethod
    def from_iu_list(cls, self, iu_list):
        """Initializes the update message with a list of tuples containing the update
        type and incremental units in the format (IncrementalUnit, UpdateType)

        Args:
            iu_list (list): A list of IncrementalUnit-UpdateType-tuples in the format
                (IncrementalUnit, UpdateType) that will be added to the update message.
        """
        um = UpdateMessage()
        um.add_ius(iu_list)
        return um

    def __iter__(self):
        return self

    def __next__(self):
        self._counter += 1
        if self._counter == len(self._msgs):
            self._counter = -1
            raise StopIteration
        return self._msgs[self._counter]

    def add_iu(self, iu, update_type, strict_update_type=True):
        """Adds an incremental unit to the update message with the given update type.

        If a single UpdateType-IncrementalUnit-Tuple raises a TypeError or a ValueError,
        none of the units will be added to the update message.

        Args:
            iu (IncrementalUnit): The incremental unit to be added to the update message
            update_type (UpdateType): The type of the update the should be associated
                with the incremental unit.
            strict_update_type (bool): Whether the update type should be checked and
                converted to type UpdateType. If the given argument is not of type
                UpdateType or a str that can be converted to UpdateType, a ValueError
                is raised.

        Raises:
            TypeError: When the given incremental unit is not of type IncrementalUnit
                or if the update_type does not correspond to an update type.
            ValueError: When the given udpate type is not a valid update type or the
                given argument cannot be converted to an UpdateType. Only applies if the
                strict_update_type flag is set.
        """
        if not isinstance(iu, IncrementalUnit):
            raise TypeError("IU is of type %s but should be IncrementalUnit" % type(iu))
        if strict_update_type and not isinstance(update_type, UpdateType):
            update_type = UpdateType(update_type)
        self._msgs.append((iu, update_type))

    def add_ius(self, iu_list, strict_update_type=True):
        """Adds a list of incremental units and according update types to the update
        message.

        Args:
            iu_list (list): A list containing tuples of update types and incremental
            units in the format (IncrementalUnit, UpdateType).
            strict_update_type (bool): Whether the update type should be checked and
                converted to type UpdateType. If the given argument is not of type
                UpdateType or a str that can be converted to UpdateType, a ValueError
                is raised.

        Raises:
            TypeError: When the given incremental unit is not of type IncrementalUnit
                or if the update_type does not correspond to an update type.
            ValueError: When the given udpate type is not a valid update type or the
                given argument cannot be converted to an UpdateType. Only applies if the
                strict_update_type flag is set.
        """
        for update_type, iu in iu_list:
            if not isinstance(iu, IncrementalUnit):
                raise TypeError(
                    "IU is of type %s but should be IncrementalUnit" % type(iu)
                )
            if strict_update_type and not isinstance(update_type, UpdateType):
                UpdateType(update_type)
        for update_type, iu in iu_list:
            if strict_update_type and not isinstance(update_type, UpdateType):
                update_type = UpdateType(update_type)
            self._msgs.append((iu, update_type))

    def has_valid_ius(self, iu_classes):
        """Checks whether the IUs in this update message are all of the type provided in
        the ius argument.

        Args:
            iu_classes (list or class): A list of incremental unit classes or a
                single incremental unit class that should be checked against.

        Returns:
            bool: Returns true if all the incremental unit in the udpate message are
                instances of the iu_classes given in the parameter.
        """
        if iu_classes is None:
            return False
        if not isinstance(iu_classes, list):
            iu_classes = [iu_classes]

        for iu in self.incremental_units():
            if not isinstance(iu, tuple(iu_classes)):
                return False
        return True

    def update_types(self):
        """A generator that iterates of all the update types of the update message,
        ignoring the incemental units
        """
        for _, ut in self._msgs:
            yield ut

    def incremental_units(self):
        """A generator that iterates of all the incremental units of the update message,
        ignoring the update types.
        """
        for iu, _ in self._msgs:
            yield iu

    def set_processed(self, module):
        """Sets all the incremental units of the update message as processed by the
        module that is given

        Args:
            module (IncrementalModule): The module that has processed the incremental
            units of this update message."""
        for iu in self.incremental_units():
            iu.set_processed(module)


class AbstractModule:
    """An abstract module that is able to incrementally process data."""

    EVENT_PROCESS_IU = "process_iu"
    EVENT_PROCESS_UPDATE_MESSAGE = "process_update_message"
    EVENT_SUBSCRIBE = "subscribe"
    EVENT_START = "start"
    EVENT_STOP = "stop"

    QUEUE_TIMEOUT = 0.01
    """Timeout in seconds for the incremental queues as not to block processing."""

    @staticmethod
    def name():
        """Return the human-readable name of the module.

        Returns:
            str: A string containing the name of the module
        """
        raise NotImplementedError()

    @staticmethod
    def description():
        """Return the human-readable description of the module.

        Returns:
            str: A string containing the description of the module
        """
        raise NotImplementedError()

    @staticmethod
    def input_ius():
        """Return the list of IU classes that may be processed by this module.

        If an IU is passed to the module that is not in this list or a subclass
        of this list, an error is thrown when trying to process that IU.

        Returns:
            list: A list of classes that this module is able to process.
        """
        raise NotImplementedError()

    @staticmethod
    def output_iu():
        """Return the class of IU that this module is producing.

        Returns:
            class: The class of IU this module is producing.
        """
        raise NotImplementedError()

    def get_init_arguments(self):
        """Returns the arguments of the init function to create the current
        instance of the Module.

        Returns:
            dict: A dictionary containing all the necessary arguments to create
            the current instance of the module.
        """
        d = {}
        valid_types = (int, float, bool, str, dict)  # Only serializable types.
        for k, v in self.__dict__.items():
            if isinstance(v, valid_types):
                d[k] = v
        return d

    def __init__(self, queue_class=IncrementalQueue, meta_data={}, **kwargs):
        """Initialize the module with a default IncrementalQueue.

        Args:
            queue_class (IncrementalQueue): A queue class that should be used
                instead of the standard queue class. If the given object does
                not inherit from IncrementalQueue, the standard IncrementalQueue
                is used.
            meta_data (dict): A dict with meta data about the module. This may
                be coordinates of the visualization of this module or other
                auxiliary information.
        """
        self._right_buffers = []
        self._is_running = False
        self._previous_iu = None
        self._left_buffers = []
        self.mutex = threading.Lock()
        self.events = {}

        self.current_input = []
        self.current_output = []

        self.meta_data = {}
        if meta_data:
            self.meta_data = meta_data

        self.queue_class = IncrementalQueue
        if issubclass(queue_class, IncrementalQueue):
            self.queue_class = queue_class

        self.iu_counter = 0

    def revoke(self, iu, remove_revoked=True):
        """Revokes an IU form the list of the current_input or current_output, depending
        on in which list it is found.

        Args:
            iu (IncrmentalUnit): The incremental unit to revoke.
            remove_revoked (bool): Whether the revoked incremental unit should be
                deleted from the current_input or current_output list or if only the
                revoked flag should be set.
        """
        if iu in self.current_input:
            self.current_input[self.current_input.index(iu)].revoked = True
            if remove_revoked:
                self.current_input.remove(iu)
        if iu in self.current_output:
            self.current_output[self.current_output.index(iu)].revoked = True
            if remove_revoked:
                self.current_output.remove(iu)

    def commit(self, iu):
        """Sets an IU as committed from the list of the current_input or current_output,
        depending on where it is found.

        Args:
            iu (IncrementalUnit): The incremental unit to set as committed.
        """
        if iu in self.current_input:
            self.current_input[self.current_input.index(iu)].committed = True
        if iu in self.current_output:
            self.current_output[self.current_output.index(iu)].committed = True

    def input_committed(self):
        """Checks whether all IUs in the input are committed.

        Returns:
            bool: True when all IUs in the current_input is committed, False otherwise.
        """
        for ciu in self.current_input:
            if not ciu.committed:
                return False
        return True

    def add_left_buffer(self, left_buffer):
        """Add a new left buffer for the module.

        This method stops the execution of the module pipeline if it is running.

        Args:
            left_buffer (IncrementalQueue): The left buffer to add to the
                module.
        """
        if not left_buffer or not isinstance(left_buffer, IncrementalQueue):
            return
        if self._is_running:
            self.stop()
        self._left_buffers.append(left_buffer)

    def remove_left_buffer(self, left_buffer):
        """Remove a left buffer from the module.

        This method stops the execution of the module pipeline if it is running.

        Args:
            left_buffer (IncrementalQueue): The left buffer to remove from the
                module.
        """
        if self._is_running:
            self.stop()
        self._left_buffers.remove(left_buffer)

    def left_buffers(self):
        """Returns the list of left buffers of the module.

        Returns:
            list: The left buffers of the module.
        """
        return list(self._left_buffers)

    def add_right_buffer(self, right_buffer):
        """Add a new right buffer for the module.

        This method stops the execution of the module pipeline if it is running.

        Args:
            right_buffer (IncrementalQueue): The right buffer to add to the
                module.
        """
        if not right_buffer or not isinstance(right_buffer, IncrementalQueue):
            return
        if self._is_running:
            self.stop()
        self._right_buffers.append(right_buffer)

    def remove_right_buffer(self, right_buffer):
        """Remove a right buffer from the module.

        This method stops the execution of the module pipeline if it is running.

        Args:
            right_buffer (IncrementalQueue): The right buffer to remove from the
                module.
        """
        if self._is_running:
            self.stop()
        self._right_buffers.remove(right_buffer)

    def right_buffers(self):
        """Return the right buffers of the module.

        Note that the returned list is only a shallow copy. Modifying the list
        does not alter the internal state of the module (but modifying the
        queues in that list does).

        Returns:
            list: A list of the right buffers, each queue corresponding to an
            input of another module.
        """
        return list(self._right_buffers)

    def append(self, update_message):
        """Append an update message to all queues.

        If update_message is None or there are no IUs in the update message, the method
        returns without doing anything.

        Args:
            update_message (UpdateMessage): The update message that should be added to
                all output queues. May be None.
        """
        if not update_message:
            return
        if not isinstance(update_message, UpdateMessage):
            raise TypeError(
                "Update message is of type %s but should be UpdateMessage"
                % type(update_message)
            )
        for q in self._right_buffers:
            q.put(copy.copy(update_message))

    def subscribe(self, module, q=None):
        """Subscribe a module to the queue.

        It returns a queue where the IUs for that module are placed. The queue
        is not shared with other modules. By default this method creates a new
        queue, but it may use an alternative queue given in parameter 'q'.

        Args:
            module (AbstractModule): The module that wants to subscribe to the
                output of the module.
            q (IncrementalQueue): A optional queue that is used. If q is None,
                the a new queue will be used"""
        if not q:
            self.event_call(self.EVENT_SUBSCRIBE, {"module": module})
            q = self.queue_class(self, module)
            module.add_left_buffer(q)
        self._right_buffers.append(q)
        return q

    def remove_from_rb(self, module):
        """Removes the connection to a module from the right buffers.

        This method removes all queues between this module and the given module
        from the right buffer of this module and the left buffer of the given
        module.
        This method stops the execution of the module.

        Args:
            module: A module that is subscribed to this module
        """
        if self._is_running:
            self.stop()
        # We get a copy of the buffers because we are mutating it
        rbs = self.right_buffers()
        for buffer in rbs:
            if buffer.consumer == module:
                buffer.remove()

    def remove_from_lb(self, module):
        """Removes the connection to a module from the left buffers.

        This method removes all queues between this module and the given module
        from the left buffer of this module and the right buffer of the given
        module.
        This method stops the execution of the module.

        Args:
            module: A module that this module is subscribed to
        """
        if self._is_running:
            self.stop()
        # We get a copy of the buffers because we are mutating it
        lbs = self.left_buffers()
        for buffer in lbs:
            if buffer.producer == module:
                buffer.remove()

    def remove(self):
        """Removes all connections to all modules.

        This methods removes all queues from the left buffer and right buffer.
        The queues are also removed from the buffers of the connected modules.
        This method can be used to remove a module completely from a network.

        This method stops the execution of the module.
        """
        if self._is_running:
            self.stop()
        lbs = self.left_buffers()
        rbs = self.right_buffers()
        for buffer in lbs:
            buffer.remove()
        for buffer in rbs:
            buffer.remove()

    def process_update(self, update_message):
        """Processes the update message given and returns a new update message that can
        be appended to the output queues.

        Note that the incremental units in the update message that is returned should be
        created by the create_iu method so that they have correct references to the
        previous incremental units generated by this module and the IUs that they are
        based on.

        It is important that the process_update method processes the update messages
        in a timely manner (in regards to the production of update messages from the
        preceding module) so that the incremental queues do not overflow.

        Args:
            update_message (UpdateMEssage): The update message that should be processed
                by the module.

        Returns:
            UpdateMessage: An update message that is produced by this module based
            on the incremental units that were given. May be None.
        """
        raise NotImplementedError()

    def _run(self):
        self.prepare_run()
        self._is_running = True
        while self._is_running:
            for buffer in self._left_buffers:
                with self.mutex:
                    try:
                        update_message = buffer.get(timeout=self.QUEUE_TIMEOUT)
                    except queue.Empty:
                        update_message = None
                    if update_message:
                        if not update_message.has_valid_ius(self.input_ius()):
                            raise TypeError("This module can't handle this type of IU")
                        output_message = self.process_update(update_message)
                        update_message.set_processed(self)
                        for input_iu in update_message.incremental_units():
                            self.event_call(self.EVENT_PROCESS_IU, {"iu": input_iu})
                        self.event_call(
                            self.EVENT_PROCESS_UPDATE_MESSAGE,
                            {"update_message": update_message},
                        )
                        if output_message:
                            if output_message.has_valid_ius(self.output_iu()):
                                self.append(output_message)
                            else:
                                raise TypeError(
                                    "This module should not produce IUs of this type."
                                )
        self.shutdown()

    def is_valid_input_iu(self, iu):
        """Return whether the given IU is a valid input IU.

        Valid is defined by the list given by the input_ius function. The given
        IU must be one of the types defined in that list or be a subclass of it.

        Args:
            iu (IncrementalUnit): The IU to be checked.

        Raises:
            TypeError: When the given object is not of type IncrementalUnit.

        Returns:
            bool: Whether the given iu is a valid one for this module.
        """
        if not isinstance(iu, IncrementalUnit):
            raise TypeError("IU is of type %s but should be IncrementalUnit" % type(iu))
        for valid_iu in self.input_ius():
            if isinstance(iu, valid_iu):
                return True
        return False

    def setup(self):
        """This method is called before the module is run. This method can be
        used to set up the pipeline needed for processing the IUs.

        However, after the setup method is called, the module may not
        immediately be run. For code that should be executed immediately before
        a module is run use the `prepare_run` method.
        """
        pass

    def prepare_run(self):
        """A method that is executed just before the module is being run.

        While this method may seem similar to `setup`, it is called immediately
        before the run routine. This method may be used in producing modules to
        initialize the generation of output IUs. Other than the `setup` method,
        this method makes sure that other modules in the network are also
        already setup.
        """
        pass

    def shutdown(self):
        """This method is called before the module is stopped. This method can
        be used to tear down the pipeline needed for processing the IUs."""
        pass

    def run(self, run_setup=True):
        """Run the processing pipeline of this module in a new thread. The
        thread can be stopped by calling the stop() method.

        Args:
            run_setup (bool): Whether or not the setup method should be executed
            before the thread is started.
        """
        if run_setup:
            self.setup()
        for q in self.right_buffers():
            with q.mutex:
                q.queue.clear()
        t = threading.Thread(target=self._run)
        t.start()
        self.event_call(self.EVENT_START)

    def stop(self, clear_buffer=True):
        """Stops the execution of the processing pipeline of this module at the
        next possible point in time. This may be after the next incoming IU is
        processed."""
        self._is_running = False
        if clear_buffer:
            for buffer in self.right_buffers():
                while not buffer.empty():
                    buffer.get()
        self.event_call(self.EVENT_STOP)

    def create_iu(self, grounded_in=None):
        """Creates a new Incremental Unit that contains the information of the
        creator (the current module), the previous IU that was created in this
        module and the iu that it is based on.

        Do not discard (as in not using) any IU that was created by this method,
        because it will alreade have been introduced into the chain of IUs of
        this module!

        Args:
            grounded_in (IncrementalUnit): The incremental unit that the new
                unit is based on. May be None.

        Returns:
            IncrementalUnit: A new incremental unit with correct pointer to
            unit it is grounded in and to the previous IU that was generated by
            this module.
        """
        new_iu = self.output_iu()(
            creator=self,
            iuid=f"{hash(self)}:{self.iu_counter}",
            previous_iu=self._previous_iu,
            grounded_in=grounded_in,
        )
        self.iu_counter += 1
        self._previous_iu = new_iu
        return new_iu

    def latest_iu(self):
        """Provides reading access to the latest incremental unit that was
        produced by this module.

        Thus, the information received by this method might be out of date or
        completely wrong (in case where a not yet initialized IU is returned).
        The iu returned should not be modified in any way, because it could
        still be processed by a module.

        Return:
            (IncrementalUnit): The latest IU that was produced by the module.
        """
        return self._previous_iu

    def __repr__(self):
        return self.name()

    def event_subscribe(self, event_name, callback):
        """
        Subscribe a callback to an event with the given name. If tge event name
        is "*", then the callback will be called after every event.

        The callback function is given three arguments: the module that
        triggered the event (AbstractModule), the name of the event (str) and a
        dict (dict) that may contain data relevant to the event.

        Args:
            event_name (str): The name of the event to subscribe to
            callback (function): A function that is called once the event occurs
        """
        if not self.events.get(event_name):
            self.events[event_name] = []
        self.events[event_name].append(callback)

    def event_call(self, event_name, data={}):
        """
        Calls all callback functions that are subscribed to the given event
        name with some data attached to it. The data is optional but should stay
        consistent with each call of the same event.

        If * is passed as the event name, no callback function is called.

        Event name should be a unique identifier to the event. "*" is not
        allowed as an event name.

        Args:
            event_name (str): The name of the event (not "*")
            data (dict): Optionally some data that is relevant to the event.
        """
        if data is None:
            data = {}
        if event_name == "*":
            return
        if self.events.get(event_name):
            for callback in self.events[event_name]:
                threading.Thread(target=callback, args=(self, event_name, data)).start()
        if self.events.get("*"):
            for callback in self.events["*"]:
                threading.Thread(target=callback, args=(self, event_name, data)).start()


class AbstractProducingModule(AbstractModule):
    """An abstract producing module that is able to incrementally process data.

    The producing module has no input queue and thus does not wait for any
    input. The producing module is called continously and may return new output
    when it becomes available.
    """

    @staticmethod
    def name():
        raise NotImplementedError()

    @staticmethod
    def description():
        raise NotImplementedError()

    @staticmethod
    def input_ius():
        return []

    @staticmethod
    def output_iu():
        raise NotImplementedError()

    def __init__(self, queue_class=IncrementalQueue, **kwargs):
        super().__init__(queue_class=IncrementalQueue, **kwargs)

    def _run(self):
        self.prepare_run()
        self._is_running = True
        while self._is_running:
            with self.mutex:
                output_message = self.process_update(None)
                if output_message:
                    if output_message.has_valid_ius(self.output_iu()):
                        self.append(output_message)
                    else:
                        raise TypeError(
                            "This module should not produce IUs of this type."
                        )
        self.shutdown()

    def process_update(self, update_message):
        raise NotImplementedError()


class AbstractConsumingModule(AbstractModule):
    """An abstract consuming module that is able to incrementally process data.

    The consuming module consumes IUs but does not return any data.
    """

    @staticmethod
    def name():
        raise NotImplementedError()

    @staticmethod
    def description():
        raise NotImplementedError()

    @staticmethod
    def input_ius():
        raise NotImplementedError()

    @staticmethod
    def output_iu():
        return None

    def subscribe(self, module, q=None):
        raise ValueError("Consuming Modules do not produce any output")

    def process_update(self, update_message):
        raise NotImplementedError()


class AbstractTriggerModule(AbstractProducingModule):
    """An abstract trigger module that produces an update message once a trigger method
    is called. Unless the module is triggered no updates are produced"""

    @staticmethod
    def name():
        raise NotImplementedError()

    @staticmethod
    def description():
        raise NotImplementedError()

    @staticmethod
    def input_ius():
        return []

    @staticmethod
    def output_iu():
        raise NotImplementedError()

    def __init__(self, queue_class=IncrementalQueue, **kwargs):
        super().__init__(queue_class=IncrementalQueue, **kwargs)

    def _run(self):
        self.prepare_run()
        self._is_running = True
        while self._is_running:
            with self.mutex:
                time.sleep(0.05)
        self.shutdown()

    def process_update(self, update_message):
        return None

    def trigger(self, data={}, update_type=UpdateType.ADD):
        """The trigger method that should produce an update message and append it to the
        right buffer

        Args:
            data (dict): A dictionary with data that can be used for the trigger
            update_type (UpdateType): The update type that the IU should have. Default
                is UpdateType.ADD
        """
        raise NotImplementedError()



"""
Network Module
==============


A Module that allows for saving and loading networks from and to file as well as
starting and stopping them based on only a single module.
"""

import pickle


def load(filename: str):
    """Loads a network from file and returns a list of modules in that network.

    The connections between the module have been set according to the file.

    Args:
        filename (str): The path to the .rtc file containing a network.

    Returns:
        (list, list): A list of Modules that are connected and ready to be run
            and a list of connections between those modules.
    """
    mc_list = pickle.load(open(filename, "rb"))
    module_dict = {}
    module_list = []
    connection_list = []

    for m in mc_list[0]:
        mod = m["retico_class"](**m["args"])
        module_dict[m["id"]] = mod
        module_list.append(mod)
    for ida, idb in mc_list[1]:
        module_dict[idb].subscribe(module_dict[ida])
        connection_list.append((module_dict[idb], module_dict[ida]))

    return (module_list, connection_list)


def load_and_execute(filename):
    """Loads a network from file and runs it.

    The network is loaded via the load-Method. Before running the network, it is
    setup.

    The networks runs until some input on the console is given

    Args:
        filename (str): The path to the .rtc file containing a network.
    """
    module_list, _ = load(filename)

    for module in module_list:
        module.setup()

    for module in module_list:
        module.run(run_setup=False)

    input()

    for module in module_list:
        module.stop()


def _discover_modules(module):
    discovered_lb = []
    discovered_rbs = []
    lbs = module.left_buffers()
    for lb in lbs:
        if lb and lb.provider:
            discovered_lb.append(lb.provider)
    for rb in module.right_buffers():
        if rb and rb.consumer:
            discovered_rbs.append(rb.consumer)
    return set(discovered_lb), set(discovered_rbs)


def run(module):
    """Properly prepares and runs a network based on one module or a list of modules.

    The network is automatically discovered so that only one module of the network has
    to be given to this function for the whole network to be executed. The function
    first calls the `setup` function of each module in the network and then runs all
    modules.

    Args:
        module (Abstract Module or list): A module of the network or a list of multiple
            module of the network
    """
    m_list, _ = discover(module)

    for m in m_list:
        m.setup()

    for m in m_list:
        m.run(run_setup=False)


def stop(module):
    """Properlystops a network based on one module or a list of modules.

    The network is automatically discovered so that only one module of the network has
    to be given to this function for the whole network to be stopped.

    Args:
        module (Abstract Module or list): A module of the network or a list of multiple
            module of the network
    """
    m_list, _ = discover(module)

    for m in m_list:
        m.stop()


def discover(module):
    """Discovers all modules and connections from a single a list of modules.

    The network is automatically discovered by traversing all left and right buffers of
    the modules given. If the argment module is only a single module, the network that
    is constructed consist only of modules and connections reachable by that module. A
    segmented network needs a module of each part of the network.

    The function returns a touple containing a list of module and a list of connections
    between the modules. The connections are touples containing the providing module of
    the connection as the first element and the receiving module as the second element.

    Args:
        module (AbstractModule or list): A module of the network or a list of multiple
            modules of the network

    Returns:
        list, list: A list of modules in the first return value and
            a list of connections in the second return value.
    """
    if not isinstance(module, list):
        module = [module]
    undiscovered = set(module)
    discovered = []
    m_list = []
    c_list = []
    while undiscovered:
        current_module = undiscovered.pop()
        discovered.append(current_module)
        lbs, rbs = _discover_modules(current_module)
        for mod in lbs:
            if mod not in discovered:
                undiscovered.add(mod)
        for mod in rbs:
            if mod not in discovered:
                undiscovered.add(mod)
        m_list.append(current_module)
        for buf in current_module.right_buffers():
            c_list.append((buf.consumer, buf.provider))
    return m_list, c_list


def save(module, filename):
    """Saves a network to file given a module or a list of modules.

    The network is automatically detected by traversing all left and right
    buffers of the modules given. If the argument module is only a single
    module, the network that is being saved consists only of the module
    reachable from this module. If a network should be saved that is splitted
    into multiple parts, at least one module of each split has to be included
    into the module-list.

    Args:
        module (AbstractModule or list): A module of the network or a list of
            multiple modules of the network.
        filename (str): The path to where the network should be stored. This
            excludes the file-ending .rtc that will be automatically added by
            this function.
    """
    if not isinstance(module, list):
        module = [module]
    undiscovered = set(module)
    discovered = []
    m_list = []
    c_list = []
    while undiscovered:
        current_module = undiscovered.pop()
        discovered.append(current_module)
        lbs, rbs = _discover_modules(current_module)
        for mod in lbs:
            if mod not in discovered:
                undiscovered.add(mod)
        for mod in rbs:
            if mod not in discovered:
                undiscovered.add(mod)
        current_dict = {}
        current_dict["widget_name"] = current_module.name()
        current_dict["retico_class"] = current_module.__class__
        current_dict["args"] = current_module.get_init_arguments()
        current_dict["id"] = id(current_module)
        current_dict["meta"] = current_module.meta_data
        m_list.append(current_dict)
        for buf in current_module.right_buffers():
            c_list.append((id(buf.consumer), id(buf.provider)))
    pickle.dump([m_list, c_list], open("%s.rtc" % filename, "wb"))



"""
Dialogue Module
===============

This module contains basic dialogue functionality, like a :class:`.DialogueActIU` for
a dialogue manager.
"""

import retico_core.core.abstract as retico_core
import time


class DialogueActIU(retico_core.IncrementalUnit):
    """A Dialog Act Incremental Unit.

    This IU represents a Dialogue Act together with concepts and their
    values. In this implementation only a single act can be expressed with a
    single IU.

    Attributes:
        act (string): A representation of the current act as a string.
        concepts (dict): A dictionary of names of concepts being mapped on to
            their actual values.
    """

    @staticmethod
    def type():
        return "Dialogue Act Incremental Unit"

    def __init__(
        self,
        creator=None,
        iuid=0,
        previous_iu=None,
        grounded_in=None,
        payload=None,
        act=None,
        concepts=None,
        **kwargs
    ):
        """Initialize the DialogueActIU with act and concepts.

        Args:
            act (string): A representation of the act.
            concepts (dict): A representation of the concepts as a dictionary.
        """
        super().__init__(
            creator=creator,
            iuid=iuid,
            previous_iu=previous_iu,
            grounded_in=grounded_in,
            payload=payload,
        )
        self.act = act
        self.concepts = {}
        if concepts:
            self.concepts = concepts
        self.confidence = 0.0

    def set_act(self, act, concepts=None, confidence=1.0):
        """Set the act and concept of the IU.

        Old acts or concepts will be overwritten.

        Args:
            act (string): The act of the IU as a string.
            concepts (dict): A dictionary containing the new concepts.
            confidence (float): Confidence of the act prediction
        """
        self.act = act
        if concepts:
            self.concepts = concepts
        self.confidence = confidence
        self.payload = (act, concepts)


class DispatchableActIU(DialogueActIU):
    """A Dialogue Act Incremental Unit that can has the information if it should
    be dispatched once it has been transformed into speech.

    Attributes:
        dispatch (bool): Whether the speech resulting from this IU should be
            dispatched or not.
    """

    def __init__(self, dispatch=False, **kwargs):
        super().__init__(**kwargs)
        self.dispatch = dispatch


class EndOfTurnIU(retico_core.IncrementalUnit):
    """An incremental unit used for prediction of the end of the turn. This
    information may be used by a dialogue management module to plan next turns
    and enabling realistic turn taking.
    """

    @staticmethod
    def type():
        return "End-of-Turn Incremental Unit"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.probability = 0.0
        self.is_speaking = False

    def set_eot(self, probability=0.0, is_speaking=False):
        """Set the end-of-turn probability and a flag if the interlocutor is
        currently speaking (VAD).

        Args:
            probability (float): The probability that the turn is ending.
            is_speaking (bool): Whether or not the interlocutor is speaking.
        """
        self.is_speaking = is_speaking
        self.probability = probability


class DialogueActRecorderModule(retico_core.AbstractConsumingModule):
    """A module that writes dispatched dialogue acts to file."""

    @staticmethod
    def name():
        return "Dialogue Act Recorder Module"

    @staticmethod
    def description():
        return "A module that writes dialogue acts into a file."

    @staticmethod
    def input_ius():
        return [DialogueActIU, DispatchableActIU]

    def __init__(self, filename, separator="\t", **kwargs):
        super().__init__(**kwargs)
        self.filename = filename
        self.separator = separator
        self.txt_file = None

    def setup(self):
        self.txt_file = open(self.filename, "w")

    def prepare_run(self):
        self.start_time = time.time()

    def shutdown(self):
        if self.txt_file:
            self.txt_file.close()
            self.txt_file = None

    def process_update(self, update_message):
        for iu, ut in update_message:
            if ut != retico_core.UpdateType.ADD:
                continue
            if self.txt_file:
                self.txt_file.write("dialogue_act")
                self.txt_file.write(self.separator)
                self.txt_file.write(str(iu.creator).split(" ")[-1])
                self.txt_file.write(self.separator)
                if iu.created_at < self.start_time:
                    self.start_time = iu.created_at
                self.txt_file.write(str(int((iu.created_at - self.start_time) * 1000)))
                self.txt_file.write(self.separator)
                self.txt_file.write("-1")
                self.txt_file.write(self.separator)
                if iu.concepts.keys():
                    self.txt_file.write(iu.act + ":" + ",".join(iu.concepts.keys()))
                else:
                    self.txt_file.write(iu.act)
                if isinstance(iu, DispatchableActIU):
                    self.txt_file.write(self.separator)
                    self.txt_file.write(str(iu.dispatch))
                self.txt_file.write("\n")


class DialogueActTriggerModule(retico_core.AbstractTriggerModule):
    @staticmethod
    def name():
        return "Dialogue Act Trigger Module"

    @staticmethod
    def description():
        return "A trigger module that emits a dialogue act when triggered."

    @staticmethod
    def output_iu():
        return DispatchableActIU

    def __init__(self, dispatch=True, **kwargs):
        super().__init__(**kwargs)
        self.dispatch = True

    def trigger(self, data={}, update_type=retico_core.UpdateType.ADD):
        output_iu = self.create_iu()
        output_iu.dispatch = self.dispatch
        output_iu.set_act(data.get("act", "greeting"), data.get("concepts", {}))
        self.append(retico_core.UpdateMessage.from_iu(output_iu, update_type))



"""
Debug Module
============

This file contains general debug modules that can be used to output information from any
incremental unit.
"""


from retico_core.core import abstract, text


class DebugModule(abstract.AbstractConsumingModule):
    """A debug module that prints the IUs that are coming in."""

    @staticmethod
    def name():
        return "Debug Module"

    @staticmethod
    def description():
        return "A consuming module that displays IU infos in the console."

    @staticmethod
    def input_ius():
        return [abstract.IncrementalUnit]

    def __init__(self, print_payload_only=False):
        super().__init__()
        self.print_payload_only = print_payload_only

    def process_update(self, update_message):
        if self.print_payload_only:
            for iu, ut in update_message:
                print(ut, iu.payload)
        else:
            print(f"Debug: Update Message ({len(update_message)})")
            for i, (iu, ut) in enumerate(update_message):
                print(f"{i}: {iu} (UpdateType: {ut})")
                print("  PreviousIU:", iu.previous_iu)
                print("  GroundedInIU:", iu.grounded_in)
                print("  Age:", iu.age())
            print(f"End of Debug Message")


class CallbackModule(abstract.AbstractConsumingModule):
    """A debug module that returns the incoming update messages into a callback
    function."""

    @staticmethod
    def name():
        return "Callback Debug Module"

    @staticmethod
    def description():
        return (
            "A consuming module that calls a callback function whenever an"
            "update message arrives."
        )

    @staticmethod
    def input_ius():
        return [abstract.IncrementalUnit]

    def __init__(self, callback, **kwargs):
        """Initializes the module with a callback function that has to take one argument
        that contains the update message whenever it arrives.
        """
        super().__init__(**kwargs)
        self.callback = callback

    def process_update(self, update_message):
        self.callback(update_message)


class TextPrinterModule(abstract.AbstractConsumingModule):
    """A debug module that prints the incoming text and updates it as text IUs are
    arriving. Once an IU is committed, the next incoming text is printed in a new
    line."""

    @staticmethod
    def name():
        return "Text Printer Module"

    @staticmethod
    def description():
        return "A module that prints out and updates text."

    @staticmethod
    def input_ius():
        return [text.TextIU]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._old_text = ""

    def process_update(self, update_message):
        print("DEBUG")
        for iu, ut in update_message:
            if ut == abstract.UpdateType.ADD:
                self.current_input.append(iu)
            elif ut == abstract.UpdateType.REVOKE:
                self.revoke(iu)
            elif ut == abstract.UpdateType.COMMIT:
                self.commit(iu)
        text = " ".join(iu.text for iu in self.current_input)
        print(" " * len(self._old_text), end="\r")
        print(text, end="\r")
        self._old_text = text
        if self.input_committed():
            self.current_input = []
            print("")



#!/usr/bin/env python3

"""
Setup script.

Use this script to install the core of the retico simulation framework. Usage:
    $ python3 setup.py install

Author: Thilo Michael (uhlomuhlo@gmail.com)
"""

try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup

exec(open("retico_core/core/version.py").read())

import pathlib

here = pathlib.Path(__file__).parent.resolve()
print(here)
print(here / "retico_core/core/README.md")

long_description = (here / "retico_core/core/README.md").read_text(encoding="utf-8")
'''
config = {
    "description": "A framework for real time incremental dialogue processing.",
    "long_description": long_description,
    "long_description_content_type": "text/markdown",
    "author": "Thilo Michael",
    "author_email": "uhlomuhlo@gmail.com",
    "url": "https://github.com/retico-team/retico-core",
    "download_url": "https://github.com/retico-team/retico-core",
    "version": __version__,
    "python_requires": ">=3.6, <4",
    "keywords": "retico, framework, incremental, dialogue, dialog",
    "install_requires": ["pyaudio>=0.2.12"],
    "packages": find_packages(include=['huggingface', 'psi', 'rasa', 'sds_class', 'slim']),
    "py_modules": ["abstract", "audio", "debug", "dialogue", "network", "text", "version", "visual"],
    "name": "pacheco-core",
    "classifiers": [
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
    ],
}
'''
config = {
    "description": "A framework for real time incremental dialogue processing.",
    "long_description": long_description,
    "long_description_content_type": "text/markdown",
    "author": "Thilo Michael",
    "author_email": "uhlomuhlo@gmail.com",
    "author": "Ryan Pacheco",
    "author_email": "ryanpacheco413@u.boisestate.edu",
    "version": "0.1.04",
    "python_requires": ">=3.6, <4",
    "keywords": "retico, framework, incremental, dialogue, dialog",
    "install_requires": ["pyaudio>=0.2.12"],
    "py_modules": ["retico_core/core/abstract",
                   "retico_core/core/audio",
                   "retico_core/core/debug",
                   "retico_core/core/dialogue",
                   "retico_core/core/network",
                   "retico_core/core/text",
                   "retico_core/core/version",
                   "retico_core/core/visual"],
    "name": "retico-core-core",
    "classifiers": [
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
    ],
}

setup(**config)



"""
This module redefines the abstract classes to fit the needs of visual processing.
"""

from retico_core.core import abstract
import numpy as np
from PIL import Image

class ImageIU(abstract.IncrementalUnit):
    """An image incremental unit that receives raw image data from a source.

    Attributes:
        creator (AbstractModule): The module that created this IU
        previous_iu (IncrementalUnit): A link to the IU created before the
            current one.
        grounded_in (IncrementalUnit): A link to the IU this IU is based on.
        created_at (float): The UNIX timestamp of the moment the IU is created.
        image (bytes[]): The image of this IU
        rate (int): The frame rate of this IU
        nframes (int): The number of frames of this IU
    """

    @staticmethod
    def type():
        return "Image IU"

    def __init__(self, creator=None, iuid=0, previous_iu=None, grounded_in=None,
                 rate=None, nframes=None, image=None,
                 **kwargs):
        super().__init__(creator=creator, iuid=iuid, previous_iu=previous_iu,
                         grounded_in=grounded_in, payload=image)
        self.image = image
        self.rate = rate
        self.nframes = nframes

    def set_image(self, image, nframes, rate):
        """Sets the audio content of the IU."""
        self.image = image
        self.payload = image
        self.nframes = int(nframes)
        self.rate = int(rate)

    def get_json(self):
        payload = {}
        payload['image'] = np.array(self.payload).tolist()
        payload['nframes'] = self.nframes
        payload['rate'] = self.rate
        return payload

    def create_from_json(self, json_dict):
        self.image =  Image.fromarray(np.array(json_dict['image'], dtype='uint8'))
        self.payload = self.image
        self.nframes = json_dict['nframes']
        self.rate = json_dict['rate']

class PosePositionsIU(abstract.IncrementalUnit):
    """An image incremental unit that maintains a list of pose_landmarks (points on body) and segmentation_mask (pixel values where 1 represents human pixel and 0 represents background pixel)).

    Attributes:
        creator (AbstractModule): The module that created this IU
        previous_iu (IncrementalUnit): A link to the IU created before the
            current one.
        grounded_in (IncrementalUnit): A link to the IU this IU is based on.
        created_at (float): The UNIX timestamp of the moment the IU is created.
    """
    @staticmethod
    def type():
        return "Pose Positions IU"

    def __init__(self, creator=None, iuid=0, previous_iu=None, grounded_in=None,
                    rate=None, nframes=None, sample_width=None, raw_audio=None,
                    **kwargs):
        super().__init__(creator=creator, iuid=iuid, previous_iu=previous_iu,
                            grounded_in=grounded_in, payload=None)
        self.pose_landmarks = None
        self.segmentation_mask = None
        self.image = None

    def set_landmarks(self, image, pose_landmarks, segmentation_mask):
        "Sets landmark content of the IU"
        self.image = image
        self.payload = pose_landmarks
        self.segmentation_mask = segmentation_mask

    def get_json(self):
        payload = {}
        payload['image'] = np.array(self.payload).tolist()
        payload['pose_landmarks'] = self.pose_landmarks
        payload['segmentation_mask'] = self.segmentation_mask
        return payload

    def create_from_json(self, json_dict):
        self.image =  Image.fromarray(np.array(json_dict['image'], dtype='uint8'))
        self.pose_landmarks = json_dict['pose_landmarks']
        self.payload = self.pose_landmarks
        self.segmentation_mask = json_dict['segmentation_mask']


class DetectedObjectsIU(abstract.IncrementalUnit):
    """An image incremental unit that maintains a list of detected objects and their bounding boxes.

    Attributes:
        creator (AbstractModule): The module that created this IU
        previous_iu (IncrementalUnit): A link to the IU created before the
            current one.
        grounded_in (IncrementalUnit): A link to the IU this IU is based on.
        created_at (float): The UNIX timestamp of the moment the IU is created.
    """

    @staticmethod
    def type():
        return "Detected Objects IU"

    def __init__(self, creator=None, iuid=0, previous_iu=None, grounded_in=None,
                 rate=None, nframes=None, sample_width=None, raw_audio=None,
                 **kwargs):
        super().__init__(creator=creator, iuid=iuid, previous_iu=previous_iu,
                         grounded_in=grounded_in, payload=None)
        self.detected_objects = None
        self.num_objects = 0
        self.image = None

    def set_detected_objects(self, image, detected_objects):
        """Sets the audio content of the IU."""
        self.image = image
        self.payload = detected_objects
        self.detected_objects = detected_objects
        self.num_objects = detected_objects['num_objs']

    def get_json(self):
        payload = {}
        payload['image'] = np.array(self.payload).tolist()
        payload['detected_objects'] = self.detected_objects
        payload['num_objects'] = self.num_objects
        return payload

    def create_from_json(self, json_dict):
        self.image =  Image.fromarray(np.array(json_dict['image'], dtype='uint8'))
        self.detected_objects = json_dict['detected_objects']
        self.payload = self.detected_objects
        self.num_objects = json_dict['num_objects']

class ObjectFeaturesIU(abstract.IncrementalUnit):
    """An image incremental unit that maintains a list of feature vectors for detected objects in a scene.

    Attributes:
        creator (AbstractModule): The module that created this IU
        previous_iu (IncrementalUnit): A link to the IU created before the
            current one.
        grounded_in (IncrementalUnit): A link to the IU this IU is based on.
        created_at (float): The UNIX timestamp of the moment the IU is created.
    """

    @staticmethod
    def type():
        return "Object Features IU"

    def __init__(self, creator=None, iuid=0, previous_iu=None, grounded_in=None,
                 rate=None, nframes=None, sample_width=None, raw_audio=None,
                 **kwargs):
        super().__init__(creator=creator, iuid=iuid, previous_iu=previous_iu,
                         grounded_in=grounded_in, payload=None)
        self.object_features = None
        self.num_objects = 0
        self.image = None

    def set_object_features(self, image, object_features):
        """Sets the audio content of the IU."""
        self.image = image
        self.payload = object_features
        self.object_features = object_features
        self.num_objects = len(object_features)

    def get_json(self):
        payload = {}
        payload['image'] = np.array(self.payload).tolist()
        payload['object_features'] = self.object_features
        payload['num_objects'] = self.num_objects
        return payload

    def create_from_json(self, json_dict):
        self.image =  Image.fromarray(np.array(json_dict['image'], dtype='uint8'))
        self.object_features = json_dict['object_features']
        self.payload = self.detected_objects
        self.num_objects = json_dict['num_objects']


class HandPositionsIU(ObjectFeaturesIU):
    """An image incremental unit that maintains a list of multi_hand_landmarks (points on hand) and multi_handedness (whether each hand is left or gith).

    Attributes:
        creator (AbstractModule): The module that created this IU
        previous_iu (IncrementalUnit): A link to the IU created before the
            current one.
        grounded_in (IncrementalUnit): A link to the IU this IU is based on.
        created_at (float): The UNIX timestamp of the moment the IU is created.
    """
    @staticmethod
    def type():
        return "Hand Positions IU"
    
    def __init__(self, creator=None, iuid=0, previous_iu=None, grounded_in=None,
                 rate=None, nframes=None, sample_width=None, raw_audio=None,
                 **kwargs):
        super().__init__(creator=creator, iuid=iuid, previous_iu=previous_iu,
                         grounded_in=grounded_in, payload=None)
        self.multi_hand_landmarks = None
        self.multi_handedness = None
        self.image = None

    def set_landmarks(self, image, multi_hand_landmarks, multi_handedness):
        "Sets landmark content of the IU"
        self.image = image
        self.payload = multi_hand_landmarks
        self.multi_handedness = multi_handedness

    def payload_to_vector(self, count):
        if self.payload is None: return
        tempPayload = self.payload
        self.payload = {}
        # print(self.payload[0])
        for hand_index, hand_landmarks in enumerate(tempPayload):
            # Iterate over the detected landmarks of the hand.
            vector = []
            for landmark in hand_landmarks.landmark:
                vector.append(landmark.x)
                vector.append(landmark.y)
                vector.append(landmark.z)
            if hand_index == 0:
                self.payload['hand' + str(hand_index)] = np.array(vector)
            # print(len(vector))

            # print("vector: ", self.payload)
        # vector = []
        # if self.payload is None: return
        # print("size ", len(self.payload))
        # for landmark in self.payload[0]:
        #     print(type(landmark))
        #     print(landmark)
        #     vector.append(landmark.x)
        #     vector.append(landmark.y)
        #     vector.append(landmark.z)
        #     print(self.payload)
        # self.payload = np.array(self.payload)


    def get_json(self):
        payload = {}
        payload['image'] = np.array(self.payload).tolist()
        payload['multi_hand_landmarks'] = self.multi_hand_landmarks
        payload['multi_handedness'] = self.multi_handedness
        return payload

    def create_from_json(self, json_dict):
        self.image =  Image.fromarray(np.array(json_dict['image'], dtype='uint8'))
        self.multi_hand_landmarks = json_dict['multi_hand_landmarks']
        self.payload = self.multi_hand_landmarks
        self.multi_handedness = json_dict['multi_handedness']

class ImageCropperModule(abstract.AbstractModule):
    """A module that crops images"""

    @staticmethod
    def name():
        return "Image Cropper Module"

    @staticmethod
    def description():
        return "A module that crops images"


    @staticmethod
    def input_ius():
        return [ImageIU]

    @staticmethod
    def output_iu():
        return ImageIU

    def __init__(self, top=-1, bottom=-1, left=-1, right=-1, **kwargs):
        """
        Initialize the Webcam Module.
        Args:
            width (int): Width of the image captured by the webcam; will use camera default if unset
            height (int): Height of the image captured by the webcam; will use camera default if unset
            rate (int): The frame rate of the recording; will use camera default if unset
        """
        super().__init__(**kwargs)
        self.top =  top
        self.bottom = bottom
        self.left = left
        self.right = right

    def process_iu(self, input_iu):
        image = input_iu.image
        width, height = image.size
        left = self.left if self.left != -1 else 0
        top = self.top if self.top != -1 else 0
        right = self.right if self.right != -1 else width
        bottom = self.bottom if self.bottom != -1 else height
        image = image.crop((left, top, right, bottom)) 
        output_iu = self.create_iu(input_iu)
        output_iu.set_image(image, input_iu.nframes, input_iu.rate)
        return output_iu